import sys
from datetime import datetime
from typing import Type, Optional, Callable, List, Union, Any

from prompt_toolkit import prompt
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.shortcuts import button_dialog, input_dialog, message_dialog
from prompt_toolkit.styles import Style

dark_mode = Style.from_dict(
    {
        "dialog": "bg:#000000",
        "dialog frame.label": "bg:#ffffff #000000",
        "dialog.body": "bg:#000000 #D3D3D3",
        "dialog shadow": "bg:#000000",
        "text-area": "#000000",
    }
)


def create_repl_prompt_str(prompt_msg: str) -> str:
    """
    >>> create_repl_prompt_str("give string!")
    'give string! > '
    >>> create_repl_prompt_str("enter an int >")
    'enter an int > '
    """
    msg = prompt_msg.strip()
    if msg.endswith(">"):
        return f"{msg} "
    else:
        return f"{msg} > "


# handles the repetitive task of validating passed kwargs for prompt string for attrs
def handle_prompt(
    for_type: Union[str, Type], for_attr: Optional[str], prompt_msg: Optional[str]
) -> str:
    # if user supplied one, use that
    pmsg = prompt_msg
    if pmsg is not None:
        return pmsg
    assert for_attr is not None, "Expected 'for_attr'; an attribute name to prompt for!"
    describe: str = for_type.__name__ if isinstance(for_type, type) else str(for_type)
    return create_repl_prompt_str(f"'{for_attr}' ({describe})")


## STRING


def prompt_str(for_attr: Optional[str] = None, prompt_msg: Optional[str] = None) -> str:
    m: str = handle_prompt(str, for_attr, prompt_msg)
    return prompt(m)


## INT


class IntValidator(Validator):
    def validate(self, document):
        text = document.text
        try:
            int(text)
        except ValueError as ve:
            raise ValidationError(message=str(ve))


def prompt_int(for_attr: Optional[str] = None, prompt_msg: Optional[str] = None) -> int:
    m: str = handle_prompt(int, for_attr, prompt_msg)
    return int(prompt(m, validator=IntValidator()))


## FLOAT


class FloatValidator(Validator):
    def validate(self, document):
        text = document.text
        try:
            float(text)
        except ValueError as ve:
            raise ValidationError(message=str(ve))


def prompt_float(
    for_attr: Optional[str] = None, prompt_msg: Optional[str] = None
) -> float:
    m: str = handle_prompt(float, for_attr, prompt_msg)
    return float(prompt(m, validator=FloatValidator()))


## BOOL


def prompt_bool(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> bool:
    m: str = handle_prompt(bool, for_attr, prompt_msg)
    return button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("True", True), ("False", False)],
        style=dark_mode,
    ).run()


## DATETIME


def prompt_datetime(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
) -> datetime:
    m: str = handle_prompt(datetime, for_attr, prompt_msg)
    use_current_time = button_dialog(
        title="How to pick date?",
        text=m,
        buttons=[("Now", True), ("Describe", False)],
        style=dark_mode,
    ).run()
    if use_current_time:
        return datetime.now()
    else:
        try:
            import dateparser  # type: ignore[import]
        except ImportError as e:
            print(str(e))
            print(
                "Could not find dateparser module, run 'pip3 install dateparser' to be able to parse date strings"
            )
            sys.exit(1)
        parsed_time: Optional[datetime] = None
        while parsed_time is None:
            time_str: Optional[str] = input_dialog(
                title="Describe the datetime:",
                text="For example:\n'2 hours ago', 'noon', 'tomorrow at 10PM', 'may 30th at 8PM'",
                style=dark_mode,
            ).run()
            if time_str is None:
                print("Cancelled, using current time...")
                return datetime.now()
            parsed_time = dateparser.parse(time_str)
            if parsed_time is None:
                message_dialog(
                    title="Error",
                    text=f"Could not parse '{time_str}' into datetime...",
                    style=dark_mode,
                ).run()
        return parsed_time


## LIST/SET repeat-prompt?


def prompt_ask_another(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> bool:
    m = prompt_msg
    if m is None:
        assert (
            for_attr is not None
        ), "Expected 'for_attr'; an attribute name to prompt for!"
        m = f"Add another item to '{for_attr}'?"
    return button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("Yes", True), ("No", False)],
        style=dark_mode,
    ).run()


## Optional


def prompt_optional(
    func: Callable[[], Any],
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> Optional[Any]:
    """
    A helper to ask if the user wants to enter information for an optional.
    If the user confirms, calls func (which asks the user for input)
    """
    m: Optional[str] = prompt_msg
    if m is None:
        assert (
            for_attr is not None
        ), "Expected 'for_attr'; an attribute name to prompt for!"
        m = f"'{for_attr}' is optional. Add?"
    if button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("Add", True), ("Skip", False)],
        style=dark_mode,
    ).run():
        return func()
    else:
        return None


##  wrap some function and display the specified thrown errors as validation errors


def prompt_wrap_error(
    func: Callable[[str], Any],
    catch_errors: Optional[List[Type]] = None,
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
) -> Any:
    """
    Takes the prompt string, some function which takes the string the user
    is typing as input, and possible errors to catch.

    If the function raises one of those errors, raise it as a ValidationError instead.

    This is pretty similar to prompt_toolkit.validators.Validation.from_callable
    but it allows you to specify the error message from the callable instead.
    """
    m: str = handle_prompt(func.__name__, for_attr, prompt_msg)

    errors_to_catch: List[Type] = catch_errors or []

    class LambdaPromptValidator(Validator):
        def validate(self, document):
            text = document.text
            try:
                func(text)
            except Exception as e:
                for catchable in errors_to_catch:
                    if isinstance(e, catchable):
                        raise ValidationError(message=str(e))
                else:
                    # if the user didnt specify this as an error to catch
                    raise e

    return func(prompt(m, validator=LambdaPromptValidator()))



import json

from typing import List, Dict, Callable, Type, NamedTuple, Any, TextIO, Optional

from .serialize import serialize_namedtuple, deserialize_namedtuple, PrimitiveType


def _pretty_print(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    if "indent" not in kwargs:
        kwargs["indent"] = "    "
    return kwargs


# need to types for *one* argument for **kwargs: https://stackoverflow.com/a/37032111/9348376
def namedtuple_sequence_dumps(
    nt_items: List[NamedTuple],
    attr_serializers: Optional[Dict[str, Callable[[Any], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    **kwargs: str,
) -> str:
    """
    Dump the list of namedtuples to a JSON string
    """
    s_obj: List[Dict[str, Any]] = []
    for nt in nt_items:
        s_obj.append(serialize_namedtuple(nt, attr_serializers, type_serializers))
    return json.dumps(s_obj, **_pretty_print(kwargs))


def namedtuple_sequence_dump(
    nt_items: List[NamedTuple],
    fp: TextIO,
    attr_serializers: Optional[Dict[str, Callable[[Any], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    **kwargs: str,
) -> None:
    """
    Dump the list of namedtuples to a file-like object as JSON
    """
    # dump to string first, so JSON serialization errors dont cause data losses
    dumped: str = namedtuple_sequence_dumps(
        nt_items, attr_serializers, type_serializers, **kwargs
    )
    fp.write(dumped)


def namedtuple_sequence_loads(
    nt_string: str,
    to: Type,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
) -> List[NamedTuple]:
    """
    Load a list of namedtuples specificed by 'to' from a JSON string
    """
    loaded_obj = json.loads(nt_string)
    if not isinstance(loaded_obj, list):
        raise TypeError(
            f"{loaded_obj} is a {type(loaded_obj).__name__}, expected a top-level list from JSON source"
        )
    ds_items: List[NamedTuple] = []
    for lo in loaded_obj:
        ds_items.append(
            deserialize_namedtuple(lo, to, attr_deserializers, type_deserializers)
        )
    return ds_items


def namedtuple_sequence_load(
    fp: TextIO,
    to: Type,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
) -> List[NamedTuple]:
    """
    Load a list of namedtuples to the namedtuple specificed by 'to'
    from a file-like object containing JSON
    """
    return namedtuple_sequence_loads(
        fp.read(), to, attr_deserializers, type_deserializers
    )



import typing
import inspect
from functools import lru_cache
from datetime import datetime
from typing import (
    Optional,
    Tuple,
    Type,
    Union,
    List,
    Any,
    Set,
    Sequence,
    NamedTuple,
    Dict,
)

# A lot of these are helpers from:
# https://github.com/karlicoss/cachew/blob/f4db4a6c6609170642c6cd09d50b52ac4c0edec9/src/cachew/__init__.py#L144

# items that can serialized directly into JSON by json.dumps
# datetime is converted to UTC before storing
PRIMITIVES = {
    str: Type[str],
    int: Type[int],
    float: Type[float],
    bool: Type[bool],
    datetime: Type[datetime],
}

PrimitiveType = Union[str, int, float, bool, datetime, dict]


CONTAINERS = {
    list: Type[List],
    set: Type[Set],
}

AnyContainerType = Union[List[Any], Set[Any]]


def cache(user_function):
    'Simple lightweight unbounded cache. Sometimes called "memoize".'
    return lru_cache(maxsize=None)(user_function)


def add_to_container(container: Union[List, Set], item: Any) -> AnyContainerType:
    if isinstance(container, list):
        container.append(item)
    elif isinstance(container, set):
        container.add(item)
    else:
        raise RuntimeError(
            f"{type(container)} is not a list/set, not sure how to add to"
        )
    return container


@cache
def get_union_args(cls: Type) -> Optional[List[Type[Any]]]:
    """
    >>> get_union_args(Union[str, int])
    [<class 'str'>, <class 'int'>]
    >>> get_union_args(Optional[str])
    [<class 'str'>]
    """
    if getattr(cls, "__origin__", None) != Union:
        return None

    args: Type = cls.__args__
    arg_list: List[Type] = [e for e in args if e != type(None)]
    assert len(arg_list) > 0
    return arg_list


# an estimation, someone could always trick
# this if they really wanted to...
# this should be passed the NamedTuple object,
# not a class type
def is_namedtuple_obj(thing: Any) -> bool:
    _asdict = getattr(thing, "_asdict", None)
    return _asdict is not None and callable(_asdict)


# this should be passed a NamedTuple type (typically
# created with class(NamedTuple), not an instance
def is_namedtuple_type(thing: Type) -> bool:
    return hasattr(thing, "_fields") and issubclass(thing, tuple) and callable(thing)


def is_union(cls: Type) -> bool:
    return get_union_args(cls) is not None


@cache
def get_collection_types(cls: Type) -> Tuple[Type, Type]:
    """
    >>> from typing import List
    >>> get_collection_types(List[int])
    (<class 'list'>, <class 'int'>)
    >>> get_collection_types(Set[bool])
    (<class 'set'>, <class 'bool'>)
    """
    container_type: Type = strip_generic(cls)
    # e.g. if List[int], internal[0] == int
    internal: Sequence[Type] = typing.get_args(cls)  # requires 3.8
    assert (
        len(internal) == 1
    ), f"Expected 1 argument for {container_type}, got {len(internal)}"
    return container_type, internal[0]


@cache
def strip_optional(cls: Type) -> Tuple[Type, bool]:
    """
    >>> from typing import Optional
    >>> strip_optional(Optional[int])
    (<class 'int'>, True)
    >>> strip_optional(int)
    (<class 'int'>, False)
    >>> strip_optional(Optional[List[int]])
    (typing.List[int], True)
    """
    is_opt: bool = False

    args: Optional[List[Type[Any]]] = get_union_args(cls)
    if args is not None and len(args) == 1:
        cls = args[0]  # meh
        is_opt = True

    return (cls, is_opt)


@cache
def strip_generic(tp):
    """
    >>> from typing import List
    >>> strip_generic(List[int])
    <class 'list'>
    >>> strip_generic(str)
    <class 'str'>
    """
    GA = getattr(typing, "_GenericAlias")
    if isinstance(tp, GA):
        return tp.__origin__
    return tp


def is_primitive(cls: Type) -> bool:
    """
    Whether or not this is a supported, serializable primitive

    >>> from typing import Dict
    >>> is_primitive(int)
    True
    >>> is_primitive(str)
    True
    >>> is_primitive(float)
    True
    >>> is_primitive(bool)
    True
    >>> is_primitive(set)
    False
    """
    return cls in PRIMITIVES


def is_supported_container(cls: Type) -> bool:
    """
    >>> from typing import Dict, List, Set, Tuple
    >>> is_supported_container(List[int])
    True
    >>> is_supported_container(Set[str])
    True
    >>> is_supported_container(Tuple[int])
    False
    >>> is_supported_container(Dict[int, str])
    False
    """
    return strip_generic(cls) in CONTAINERS


@cache
def inspect_signature_dict(
    nt: Type[NamedTuple],
) -> Dict[str, Type]:
    return {
        name: param.annotation
        for name, param in inspect.signature(nt).parameters.items()
    }



from datetime import datetime
from typing import NamedTuple

from autotui.shortcuts import load_prompt_and_writeback


class Water(NamedTuple):
    at: datetime
    glass_count: float


if __name__ == "__main__":
    load_prompt_and_writeback(Water, "~/.local/share/water.json")



from pathlib import Path
from typing import (
    Callable,
    Type,
    NamedTuple,
    Union,
    Dict,
    List,
    Any,
    Optional,
)


from . import (
    AutoHandler,
    prompt_namedtuple,
    namedtuple_sequence_dumps,
    namedtuple_sequence_load,
)
from .typehelpers import PrimitiveType


def _normalize(_path: Union[Path, str]) -> Path:
    p: Path
    if isinstance(_path, str):
        p = Path(_path)
    else:
        p = _path
    return p.expanduser().absolute()


# doesnt expose underlying kwargs from namedtuple_sequence_dump
# and namedtuple_sequence_load on purpose -- so that its less
# likely its mistyped. Can always use the underlying
# functions if you'd prefer to do that.


def dump_to(
    items: List[NamedTuple],
    path: Union[Path, str],
    attr_serializers: Optional[Dict[str, Callable[[Any], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
) -> None:
    """
    Takes a list of NamedTuples (or subclasses) and a path to a file.
    Serializes the items into a string, using the attr_serializers and
    type_serializers to handle custom types if specified.

    If serialization succeeds, writes to the file.
    """
    p = _normalize(path)
    # serialize to string before opening file
    # if serialization fails, file is left alone
    nt_string: str = namedtuple_sequence_dumps(
        items, attr_serializers=attr_serializers, type_serializers=type_serializers
    )
    with p.open(mode="w") as f:
        f.write(nt_string)


# args are slightly reordered here, comapared to json.load
# to be consistent with dump_to
def load_from(
    to: Type,
    path: Union[Path, str],
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    allow_empty: bool = False,
) -> List[NamedTuple]:
    """
    Takes a type to load and a path to a file containing JSON.
    Reads from the file, and deserializes the items into a list,
    using the attr_deserializers and type_deserializers to handle
    custom types if specified.

    If allow_empty is True, this returns an empty list if the file
    didn't exist

    Returns the list of items read from the file.
    """
    p: Path = _normalize(path)
    try:
        with p.open(mode="r") as f:
            items: List[NamedTuple] = namedtuple_sequence_load(
                f,
                to,
                attr_deserializers=attr_deserializers,
                type_deserializers=type_deserializers,
            )
    except FileNotFoundError as fne:
        if allow_empty:
            return []
        else:
            raise fne
    return items


def load_prompt_and_writeback(
    to: Type,
    path: Union[Path, str],
    *,
    create_file: bool = True,
    attr_validators: Optional[Dict[str, AutoHandler]] = None,
    type_validators: Optional[Dict[Type, AutoHandler]] = None,
    attr_use_values: Optional[Dict[str, Any]] = None,
    type_use_values: Optional[Dict[Type, Any]] = None,
    attr_serializers: Optional[Dict[str, Callable[[Any], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    prompt_function: Optional[Callable[..., NamedTuple]] = None,
) -> List[NamedTuple]:
    """
    An entry point to entire library, essentially.

    Load the NamedTuples from the JSON file specified by 'path' and 'to'

    If the file doesn't exist and the create_file flag is True, ignores the
    FileNotFound error when trying to read from the file.

    After reading the NamedTuples from the file, prompts you to add a new item,
    adds that to the list it reads in, and writes back to the file

    The 'prompt_function' overrides the default 'prompt_namedtuple', if you
    want to change how to prompt/the underling functions generated by namedtuple_prompt_funcs

    accepts validators, serializers and deserializers for each step of the process.
    """
    p: Path = _normalize(path)
    # read from file
    items: List[NamedTuple] = []
    try:
        items = load_from(to, p, attr_deserializers, type_deserializers)
    except FileNotFoundError as fne:
        if not create_file:
            raise fne
        # warnings.warn(f"File at {p} didn't exist, using empty list")
    # prompt for the new item
    chosen_prompt_function = prompt_function or prompt_namedtuple
    new_item: NamedTuple = chosen_prompt_function(
        to,
        attr_validators=attr_validators,
        type_validators=type_validators,
        attr_use_values=attr_use_values,
        type_use_values=type_use_values,
    )
    items.append(new_item)
    # dump back to file
    dump_to(items, p, attr_serializers, type_serializers)
    return items
